from pandas import DataFrame


def _has_dataframe(obj):
	if isinstance(obj, dict):
		for value in obj.values():
			if isinstance(value, DataFrame):
				return True
			if isinstance(value, (dict, list, tuple, set)):
				if _has_dataframe(value):
					return True
		return False
	elif isinstance(obj, (list, tuple, set)):
		for element in obj:
			if isinstance(element, DataFrame):
				return True
			if isinstance(element, (dict, list, tuple, set)):
				if _has_dataframe(element):
					return True
		return False
	elif isinstance(obj, DataFrame):
		return True

	else:
		return False


TAB = ' ' * 4


def _display_dataframe_container(obj, display_function, have_display, obj_has_dataframe, depth=0):

	if obj_has_dataframe is None:
		obj_has_dataframe = _has_dataframe(obj)

	if not obj_has_dataframe:
		print(TAB * depth, obj, sep='')

	else:
		if display_function is None and have_display is None:
			try:
				from IPython.core.display import display
				have_display = True
			except ImportError:
				display = None
				have_display = False
		elif display_function is None and not have_display:
			display = None

		else:
			display = display_function

		if isinstance(obj, DataFrame):
			if have_display:
				display(obj)
			else:
				print(TAB * depth, obj, sep='')

		elif isinstance(obj, dict):
			for key, value in obj.items():
				value_has_dataframe = _has_dataframe(value)

				if value_has_dataframe:
					print(TAB * depth, f'{key}:', sep='')
					_display_dataframe_container(
						obj=value, display_function=display, have_display=have_display,
						obj_has_dataframe=value_has_dataframe, depth=depth + 1
					)
				else:
					print(TAB * depth, f'{key}: {value}')

		elif isinstance(obj, (list, set, tuple)):
			for element in obj:
				_display_dataframe_container(
					obj=element, display_function=display, obj_has_dataframe=None, have_display=have_display,
					depth=depth + 1
				)

		else:
			_display_dataframe_container(
				obj=obj, display_function=display_function, have_display=have_display,
				obj_has_dataframe=None, depth=depth + 1
			)


def display_dataframe_container(obj):
	return _display_dataframe_container(
		obj=obj, display_function=None, obj_has_dataframe=None, have_display=None, depth=0
	)


class JupyterContainer:
	def display(self, p=None):
		display_dataframe_container(self)

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text('Validation Dictionary')
		else:
			self.display(p=p)


class JupyterDictionary(JupyterContainer, dict):
	pass


class JupyterList(JupyterContainer, list):
	pass


class JupyterTuple(JupyterContainer, tuple):
	pass


class JupyterSet(JupyterContainer, set):
	pass



import pickle
import os
my_path = os.path.abspath(os.path.dirname(__file__))
data_dir = os.path.join(my_path, '../data_files')
x11_path = os.path.join(data_dir, 'x11_colours.pickle')
svg_path = os.path.join(data_dir, 'svg_colours.pickle')
colour_schemes_path = os.path.join(data_dir, 'colour_schemes.pickle')

with open(file=x11_path, mode='rb') as x11_file:
	hexadecimal_to_x11 = pickle.load(file=x11_file)

with open(file=svg_path, mode='rb') as svg_file:
	hexadecimal_to_svg = pickle.load(file=svg_file)

hexadecimal_to_name = {}
name_to_hexadecimal = {}
for hexadecimal, name in hexadecimal_to_svg.items():
	if hexadecimal not in hexadecimal_to_name:
		hexadecimal_to_name[hexadecimal] = name
	if name not in name_to_hexadecimal:
		name_to_hexadecimal[name] = hexadecimal

for hexadecimal, name in hexadecimal_to_x11.items():
	if hexadecimal not in hexadecimal_to_name:
		hexadecimal_to_name[hexadecimal] = name
	if name not in name_to_hexadecimal:
		name_to_hexadecimal[name] = hexadecimal

with open(file=colour_schemes_path, mode='rb') as colour_schemes_file:
	colour_schemes = pickle.load(file=colour_schemes_file)



from pandas import DataFrame


def move_columns(data, to, columns=None):
	"""
	brings the columns to the front of dataframe
	:type data: DataFrame
	:param str to: 'front' brings to front, 'back' sends to back
	:type columns: list[str] or NoneType
	:rtype: DataFrame or NoneType
	"""
	columns = columns or []
	if isinstance(columns, str):
		columns = [columns]
	other_columns = [column for column in data.columns if column not in columns]
	if to == 'front':
		return data[columns + other_columns]
	elif to == 'back':
		return data[other_columns + columns]


def bring_to_front(data, columns=None):
	"""
	brings the columns to the front of dataframe
	:type data: DataFrame
	:type columns: list[str] or NoneType
	:rtype: DataFrame or NoneType
	"""
	return move_columns(data=data, to='front', columns=columns)


def send_to_back(data, columns=None):
	"""
	brings the columns to the front of dataframe
	:type data: DataFrame
	:type columns: list[str] or NoneType
	:rtype: DataFrame or NoneType
	"""
	return move_columns(data=data, to='back', columns=columns)



from .ProgressBar import ProgressBar


class Iterable:
	def __init__(self, inner, progress_bar=None, text='', echo=1, echo_items=False):
		"""
		:type inner: iterable
		:type progress_bar: ProgressBar
		:type text: str
		:type echo: bool or int or ProgressBar
		:type echo_items: bool or list[str]
		"""
		self._idx = 0
		try:
			self._total = len(inner)
			self._inner = inner
		except TypeError:
			self._inner = list(inner)
			self._total = len(self._inner)
		self._text = text
		self._echo_items = echo_items
		self._echo = echo

		if progress_bar is None:
			self._progress_bar = ProgressBar(total=self._total, echo=echo)
		else:
			self._progress_bar = progress_bar
			self._progress_bar._total = len(self._inner)

	def __iter__(self):
		return self

	def __next__(self):
		self._idx += 1
		progress_amount = self._idx - 1
		try:
			result = self._inner[self._idx - 1]
			if self._echo:
				text = f'{self._text} {result}' if self._echo_items else self._text
				self._progress_bar.show(amount=progress_amount, text=text)
			return result
		except IndexError:
			self._idx = 0
			self._progress_bar.show(amount=progress_amount + 1, text=self._text)
			raise StopIteration


def iterate(iterable, progress_bar=None, text='', echo=1, echo_items=False):
	"""
	:type iterable: Iterable
	:type progress_bar: ProgressBar
	:type text: str
	:type echo: bool or int or ProgressBar
	:type echo_items: bool or list[str]
	:rtype: Iterable
	"""
	return Iterable(inner=list(iterable), progress_bar=progress_bar, text=text, echo=echo, echo_items=echo_items)



from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from pyspark.ml.functions import vector_to_array
from pyspark.sql.types import FloatType


def disassemble(data, column, names=None, prefix=None, drop=True):
	"""
	disassembles a column of vectors into multiple columns
	:type data: DataFrame
	:type column: str
	:type names: NoneType or list[str]
	:param names: names of new columns, if not provided the prefix or column will be used
	:type prefix: NoneType or str
	:param prefix: a prefix to be used instead of names when names is not provided
	for example, if prefix='feature_' the new columns will be 'feature_1', 'feature_2', etc.
	if prefix is not provided either, the column name will be used with an underscore after
	:type drop: bool
	:param drop: if True, the original column of vectors will be dropped
	:rtype: DataFrame
	"""
	original_columns = data.columns
	if drop:
		original_columns = [col for col in original_columns if col != column]
	disassembled = data.withColumn('_array_', vector_to_array(column))
	length = len(disassembled.first()['_array_'])
	if names is None:
		if prefix is None:
			prefix = f'{column}_'
		names = [f'{prefix}{i}' for i in range(1, length + 1)]
	else:
		if len(names) != length:
			raise ValueError(f'names should be of size {length}')

	@f.udf(FloatType())
	def _get_item(array, i):
		return array[i]

	return disassembled.select(
		*original_columns,
		*[_get_item('_array_', f.lit(i)).alias(name) for i, name in enumerate(names)]
	)




from copy import deepcopy
from numpy import where, minimum, maximum


def imputate_with_regression(
		data, regressor, based_on_col, group_by = None,
		cols=None, except_cols=[], inplace=False,
		limit_min = False,
		limit_max = False,
		global_min = None,
		global_max = None
):

	if inplace:
		new_data = data
	else:
		new_data = data.copy()

	if cols is None:
		cols = new_data.columns

	if group_by is None:

		for col in cols:
			if col not in except_cols and col != based_on_col and col and new_data[col].isnull().values.any():
				available_data = data[data[col].isnull()==False]

				this_regressor = deepcopy(regressor)
				try:
					this_regressor.train(X=available_data[[based_on_col]], y=available_data[col], echo=False)
					pred = this_regressor.predict(data=data[[based_on_col]], echo=False)
					if limit_min:
						if global_min is not None:
							the_min = global_min
						else:
							the_min = available_data[col].min()
						pred = maximum(pred, the_min)

					if limit_max:
						if global_max is not None:
							the_max = global_max
						else:
							the_max = available_data[col].max()
						pred = minimum(pred, the_max)

					new_data[col] = where(data[col].isnull(), pred, data[col])
				except:
					pass

	else:
		return data.groupby(group_by).apply(
			lambda x: imputate_with_regression(
				data=x, based_on_col=based_on_col, cols=cols, except_cols=except_cols,
				inplace=False, regressor=regressor, limit_min=limit_min, limit_max=limit_max,
				global_min=global_min, global_max=global_max
			)
		).reset_index(drop=True)

	return new_data


def namespace_has(namespace, obj_type, obj_id):
	return hasattr(namespace, f'{obj_type}_{obj_id}')


def namespace_has_data(namespace, data_id):
	return namespace_has(namespace=namespace, obj_type='data', obj_id=data_id)


def add_obj_to_namespace(namespace, obj_type, obj_id, obj, overwrite=False):
	if namespace_has(namespace=namespace, obj_type=obj_type, obj_id=obj_id):
		if not overwrite:
			raise ValueError(f'{obj_type} {obj_id} already exists in the namespace')

	setattr(namespace, f'{obj_type}_{obj_id}', obj)


def get_data_from_namespace(namespace, data_id):
	return get_obj_from_namespace(namespace=namespace, obj_type='data', obj_id=data_id)


def get_obj_from_namespace(namespace, obj_type, obj_id):
	return getattr(namespace, f'{obj_type}_{obj_id}')



from .get_elapsed import get_elapsed
from .get_time import get_now


class Job:
	def __init__(self):
		self._start = None
		self._end = None

	def start(self):
		if self._start is None:
			self._start = get_now()
		else:
			raise RuntimeError('Job has already started!')

	def end(self):
		if self._start is None:
			raise RuntimeError('Job has not started yet!')
		elif self._end is None:
			self._end = get_now()
		else:
			raise RuntimeError('Job has already ended!')

	@property
	def start_time(self):
		if self._start is None:
			raise RuntimeError('Job has not started yet!')
		return self._start

	@property
	def end_time(self):
		if self._end is None:
			raise  RuntimeError('Job has not ended yet!')
		return self._end

	def get_elapsed(self, unit='ms'):
		return get_elapsed(start=self.start_time, end=self.end_time, unit=unit)



from pandas import DataFrame
from math import ceil


def split_data(data, split_size=None, num_splits=None):
	"""
	:type data: DataFrame
	:type split_size: int
	:type num_splits: int
	:rtype: generator
	"""
	if split_size is None and num_splits is None:
		raise ValueError('either size or num_splits should be provided')
	if split_size is None:
		split_size = len(data) / num_splits
	if num_splits is None:
		num_splits = ceil(len(data) / split_size)

	for i in range(num_splits):
		start = round(i * split_size)
		end = min(round((i + 1) * split_size), len(data))
		this_data = data.iloc[start:end]
		yield this_data



import re


def remove_non_alphanumeric(s, replace_with='_', keep_underscore=True):
	s = str(s)
	if keep_underscore:
		return re.sub(r'[\W]+', replace_with, s, flags=re.UNICODE)
	else:
		return re.sub(r'[\W_]+', replace_with, s, flags=re.UNICODE)



from numpy import random
import pandas as pd


def create_data(num_rows, num_x_columns, noise):
    columns = [f'x_{i + 1}' for i in range(num_x_columns)]
    df = pd.DataFrame({
        column: random.normal(1, 100, num_rows)
        for column in columns
    })
    df['y'] = 1
    for i, column in enumerate(columns):
        df['y'] = df['y'] + (df[column] ** 2 % (i + 1))
    df['y'] = df['y'] + 1
    df['y'] = df['y'] * df['y'] / 1e5
    df['y'] += random.normal(0, noise, num_rows)
    return df



from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from pyspark.sql.types import StringType
from memoria import smart_hash


class NullReplacer:
	def __init__(self):
		self._mappers = {}

	@staticmethod
	def _get_null_replaced_column_name(column_name):
		if column_name.startswith('_') and column_name.endswith('_null_replaced_'):
			raise ValueError(f'"{column_name}" matches the null replaced column name pattern!')
		return f'_{column_name}_null_replaced_'

	@staticmethod
	def _get_column_name(null_replaced_column_name):
		if null_replaced_column_name.startswith('_') and null_replaced_column_name.endswith('_null_replaced_'):
			return null_replaced_column_name[1:-len('_null_replaced_')]
		else:
			raise ValueError(f'"{null_replaced_column_name}" does not match the null replaced column name pattern!')

	def transform(self, X, columns=None):
		"""
		:param X: dataframe to be fixed
		:param columns: list of columns to be fixed by replacing nulls
		:type X: DataFrame
		:type columns: list[str]
		:rtype: DataFrame
		"""
		if columns is None:
			columns = X.columns

		@f.udf(StringType())
		def _hash(x):
			return smart_hash(x)

		result = X.select(
			*X.columns,
			*[_hash(column).alias(self._get_null_replaced_column_name(column)) for column in columns]
		)

		non_converted_columns = [col for col in X.columns if col not in columns]

		for col in columns:
			new_mapper = result.select(col, self._get_null_replaced_column_name(col)).distinct()
			if col not in self._mappers:
				self._mappers[col] = new_mapper
			else:
				self._mappers[col] = new_mapper.union(self._mappers[col]).distinct()

		return result.select(
			*non_converted_columns,
			*[f.col(self._get_null_replaced_column_name(col)).alias(col) for col in columns]
		)

	def revert(self, X):
		"""
		:type X: DataFrame
		:rtype: DataFrame
		"""
		result = X
		for col, mapper in self._mappers.items():
			if col in X.columns:
				null_replaced_column = self._get_null_replaced_column_name(col)
				result = result.withColumnRenamed(col, null_replaced_column).join(
					mapper, on=null_replaced_column, how='inner'
				).drop(null_replaced_column)
		return result



class Series:
	def __init__(self, start, end, step=None, count=None, include_start=True, include_end=True):
		self._start = start
		self._end = end

		if count is None and step is None:
			raise ValueError('either count should be given or step!')
		elif count is not None and step is not None:
			raise ValueError('only one of step and count should be given!')
		elif step is not None:
			if start <= end:
				step = abs(step)
			else:
				step = -abs(step)

		if not include_start and not include_end:
			raise ValueError('either start or end should be included!')

		self._step = step
		self._count = count
		self._include_start = include_start
		self._include_end = include_end

	def generate(self):
		if self._include_start and self._include_end:
			if self._count is None:
				num = self._start
				while num <= self._end:
					yield num
					num += self._step
			else:
				step = (self._end - self._start) / (self._count - 1)
				for i in range(self._count - 1):
					yield self._start + i * step
				yield self._end

		elif self._include_start:
			if self._count is None:
				num = self._start
				while num < self._end:
					yield num
					num += self._step
			else:
				step = (self._end - self._start) / self._count
				for i in range(self._count):
					yield self._start + i * step
		elif self._include_end:
			if self._count is None:
				count = (self._end - self._start) // self._step
				start = self._end - count * self._step
				if start != self._start:
					for i in range(count, -1, -1):
						yield self._end - i * self._step
				else:
					for i in range(count - 1, -1, -1):
						yield self._end - i * self._step
			else:
				step = (self._end - self._start) / self._count
				for i in range(1, self._count):
					yield self._start + step * i
				yield self._end
		else:
			raise ValueError('either start or end should be included!')

	def __getstate__(self):
		return self._start, self._end, self._step, self._count, self._include_start, self._include_end

	def __setstate__(self, state):
		self._start, self._end, self._step, self._count, self._include_start, self._include_end = state

	def __hashkey__(self):
		return self.__class__.__name__, self.__getstate__()

	@property
	def list(self):
		return list(self.generate())



import datetime


def get_first_day_of_month(year=None, month=None, date=None):
	"""
	:rtype: datetime.date
	"""
	if date is None and year is not None and month is not None:
		return datetime.date(year, month, 1)
	elif date is not None and year is None and month is None:
		return date.replace(day=1)
	else:
		raise RuntimeError('either year and month should be provided or date and not all!')


def get_last_day_of_month(year=None, month=None, date=None):
	"""
	:rtype: datetime.date
	"""
	if date is None:
		date = datetime.date(year, month, 28)
	else:
		date = date.replace(day=28)

	next_month = date + datetime.timedelta(days=4)
	return next_month - datetime.timedelta(days=next_month.day)




from ._select_extreme import select_max, select_min
from ._long_to_wide import long_to_wide, tall_to_wide
from ._reset_column_index import reset_column_index
from ._standardize_columns import standardize_columns
from ._fill_na import fill_na
from ._as_string import as_string
from ._as_numeric import as_int, as_numeric
from ._split_rows import split_rows
from ._select_columns import select_columns
from ._imputate_with_regression import imputate_with_regression
from ._columns_are_unique import columns_are_unique
from ._set_value_where import set_value_where
from ._join_wisely import join_wisely, join_and_keep_order
from ._min_max import min, max
from ._fast_merge import fast_merge
from ._find_duplicated_columns import find_duplicated_columns
from ._join_and_preserve_types import join_and_preserve_types
from ._bring_to_front import move_columns, bring_to_front, send_to_back
from ._create_data_from_combinations import create_data_from_combinations
from ._read_excel import read_excel
from ._drop_bad_columns import drop_bad_columns
from ._count_rows_with_missing import count_rows_with_missing, count_rows_without_missing
from ._apply_function import apply_function
from .NearestFinder import NearestFinder
from ._get_string_columns import get_string_columns
from .NullReplacer import NullReplacer
from .get_modified_zscore import get_modified_zscore, is_outlier, remove_outliers
from .union import union



import datetime as dt
from .remove_non_alphanumeric import remove_non_alphanumeric


# DAY
def add_days(date, days):
	if date is None:
		return None
	else:
		return date + dt.timedelta(days=days)


# WEEK
def add_weeks(date, weeks):
	if date is None:
		return None
	else:
		return add_days(date=date, days=weeks * 7)


# MONTH
def add_months(date, months, round='down'):
	if date is None:
		return None
	else:
		targetmonth = months + date.month
		try:
			date = date.replace(year=date.year + int(targetmonth / 12), month=(targetmonth - 1) % 12 + 1)
		except:
			# There is an exception if the day of the month we're in does not exist in the target month
			# Go to the FIRST of the month AFTER
			date = date.replace(year=date.year + int((targetmonth + 1) / 12), month=targetmonth % 12 + 1, day=1)
			if round == 'down':
				# then go back one day.
				date += dt.timedelta(days=-1)
		return date


def add_quarters(date, quarters, round='down'):
	if date is None:
		return None
	else:
		return add_months(date=date, months=quarters * 3, round=round)


# YEAR
def add_years(date, years, round='down'):
	"""Return a date that's `years` years after the date (or datetime)
	object `d`. Return the same date date (month and day) in the
	destination year, if it exists, otherwise use the following day
	(thus changing February 29 to Feb 28).

	"""
	if date is None:
		return None
	else:
		try:
			return date.replace(year=date.year + years)
		except ValueError:
			if round == 'down':
				return date.replace(year=date.year + years, day=date.day - 1)
			else:
				return date.replace(year=date.year + years, month=date.month + 1, day=1)


# TIME
def add_time(t, amount, unit='day', round='down'):
	unit = remove_non_alphanumeric(s=unit.lower(), replace_with='')
	if unit == 'day':
		return add_days(date=t, days=amount)
	if unit == 'week':
		return add_weeks(date=t, weeks=amount)
	elif unit == 'month':
		return add_months(date=t, months=amount, round=round)
	elif unit == 'year':
		return add_years(date=t, years=amount, round=round)
	elif unit == 'quarter':
		return add_quarters(date=t, quarters=amount, round=round)
	else:
		return t



from sys import stdout
from datetime import datetime
from time import sleep
import pandas as pd
from math import floor

from ...colour import Colour
_map = map

SECONDS_IN_A_DAY = 60.0 * 60.0 * 24.0


def elapsed(start_time, end_time):
	delta = end_time - start_time
	return delta.days * SECONDS_IN_A_DAY + delta.seconds + delta.microseconds / 1E6


def colour(text, text_colour='blue', background_colour=None):
	return Colour(name=text_colour).colourize(string=text, background=background_colour)


class ProgressBar:
	def __init__(
			self, total, bar_length=20, animation='clock',
			full_colour='blue', empty_colour='grey', text_colour='grey', next_line=True,
			echo=1, parent=None, disappear=False, display_wait=0.5,
			bar_characters=None
	):
		"""
		:type total: int or float or NoneType
		:type bar_length: int
		:type bar_full: str
		:type bar_empty: str
		:type animation: str
		:type full_colour: str
		:type empty_colour: str
		:type text_colour: str
		:type next_line: bool
		:type echo: int or bool
		:type parent: ProgressBar or NoneType
		:type display_wait: float or int
		"""
		self._total = total
		self._amount = None
		self._safe_total = total
		self._bar_characters = bar_characters or self.FILLING_SQUARE
		self._shown = 0
		self.amount = 0

		self._bar_length = bar_length
		self._start_time = datetime.now()
		self._animation_counter = -1
		if animation == 'vertical_bar':
			animation_clips = ['▁', '▂', '▃', '▄', '▅', '▆', '▇', '█']
			animation_clips = animation_clips + animation_clips[::-1]
		elif animation == 'ball':
			animation_clips = ['⠁', '⠂', '⠄', '⡀', '⢀', '⠠', '⠐', '⠈']
		elif animation == 'clock':
			animation_clips = ['🕐', '🕑', '🕒', '🕓', '🕔', '🕕', '🕖', '🕗', '🕘', '🕙', '🕚', '🕛']
		elif animation == 'big_ball':
			animation_clips = ['●    ', '●    ', ' ●   ', ' ●   ', '  ●  ', '   ● ', '   ● ', '    ●', '    ●']
			animation_clips = animation_clips + animation_clips[::-1]
		elif animation == 'line':
			animation_clips = ['\\', '|', '/', '-']
		else:
			animation_clips = None
		self._animation_clips = animation_clips
		self._full_colour = full_colour
		self._empty_colour = empty_colour
		self._text_colour = text_colour
		self._next_line = next_line
		self._disappear = disappear
		self._completed = False
		self._display_wait = display_wait
		self._display_time = datetime.now()

		if isinstance(echo, self.__class__):
			self._echo = echo._echo
			self._parent = echo
			self._next_line = False
		else:
			self._echo = max(echo, 0)
			self._parent = None

		if parent is not None:
			self._parent = parent
			self._next_line = False

		self._last_text = ''
		self._last_formatted_bar = ''

		if self._parent:
			self._max_lengths = self._parent._max_lengths
		else:
			self._max_lengths = {
				'animation': 0, 'elapsed': 0, 'remaining': 0, 'text': 0, 'percent': 0, 'with_colour': 0
			}

	def __sub__(self, other):
		if isinstance(other, int):
			result = self.__class__(total=self._total, full_colour='green', echo=self._echo - other, parent=self, next_line=False)
			result._amount = self._amount
			return result
		else:
			raise TypeError('subtract only works with integers')

	def set_total(self, total):
		self._total = total
		return self

	@property
	def animation(self):
		if self._animation_clips is None:
			return ''
		else:
			self._animation_counter += 1
			return self._animation_clips[self._animation_counter % len(self._animation_clips)] + ' '

	def complete(self):
		self._completed = True

	@property
	def completed(self):
		return self._completed

	@property
	def total(self):
		return self._safe_total

	@property
	def amount(self):
		return self._amount

	@amount.setter
	def amount(self, amount):
		try:
			self._amount = min(amount, self._total)
		except TypeError:
			self._amount = amount
		self._safe_total = self._total or (amount + 10)

	@property
	def percent(self):
		if self._parent:
			return self._parent.percent

		try:
			return self.amount / self.total * 100
		except ZeroDivisionError:
			if self._parent:
				return self._parent.percent
			else:
				return 0
		except TypeError:
			if self._parent:
				return self._parent.percent
			else:
				return 0

	@property
	def percent_formatted(self):
		formatted_percent = '{0: >#06.2f}'.format(float(self.percent)) + '%'
		return formatted_percent

	@property
	def elapsed_seconds(self):
		if self._parent:
			return self._parent.elapsed_seconds
		delta = datetime.now() - self._start_time
		delta_seconds = delta.seconds + delta.microseconds / 1E6
		return delta_seconds

	@property
	def remaining_seconds(self):
		if self._parent:
			return self._parent.remaining_seconds

		delta_seconds = self.elapsed_seconds

		try:
			speed = self.amount / delta_seconds

			if self.amount >= self.total:
				return 0

			else:
				if speed == 0 or speed is None:
					return None
				else:
					return (self.total - self.amount) / speed

		except ZeroDivisionError:
			return None
		except TypeError:
			return None

	@staticmethod
	def format_time(time):
		"""
		:type time: float or int
		:rtype: str
		"""
		# convert to the right unit
		if time is None:
			return ''
		else:
			if time > 3600:
				unit = 'h'
				time = time / 3600
			elif time > 60:
				unit = 'm'
				time = time / 60
			else:
				unit = 's'
				time = time

			return '{0: >#04.1f}'.format(float(time)) + unit

	FADE_CHARACTERS = [' ', '░', '▒', '▓', '█']
	LEFT_TO_RIGHT_GROWTH = ["　", "▏", "▎", "▍", "▋", "▊", "▉"]
	FOUR_SQUARES = ['　', '▖', '▞', '▛', '▉']
	FILLING_SQUARE = ['◻', '▢', '▧', '▨', '▦', '▩', '◼']

	@property
	def bar(self):
		"""
		:rtype: str
		"""

		if self.total == 0 or self.amount is None or self.total is None:
			full_part_len = 0
			partial_character = ''
		else:
			bar_length = self.amount*self._bar_length / self.total
			full_part_len = floor(bar_length)
			partial_part = bar_length - full_part_len
			if full_part_len < bar_length:
				partial_character = self._bar_characters[round(partial_part*(len(self._bar_characters)-1))]
			else:
				partial_character = ''

		empty_part_len = self._bar_length - full_part_len - len(partial_character)


		full_part_text = self._bar_characters[-1] * full_part_len + partial_character
		empty_part_text = self._bar_characters[0] * empty_part_len

		if self._full_colour is not None:
			character = '▅'  # character = self._bar_empty
			full_part = colour(text=full_part_text, text_colour=self._full_colour)
			empty_part = colour(text=empty_part_text, text_colour=self._empty_colour)
		else:
			full_part = full_part_text
			empty_part = empty_part_text

		return full_part + empty_part

	@staticmethod
	def write(string, flush=True):
		"""
		:type string: str
		"""
		stdout.write('\r'+string)
		if flush:
			stdout.flush()

	@property
	def _animation_string(self):
		animation = self.animation
		self._max_lengths['animation'] = max(self._max_lengths['animation'], len(animation))
		return colour(text=animation.ljust(self._max_lengths['animation']), text_colour=self._empty_colour)

	@property
	def _elapsed_time_string(self):
		elapsed_text = f'e:{self.format_time(self.elapsed_seconds)} '
		self._max_lengths['elapsed'] = max(self._max_lengths['elapsed'], len(elapsed_text))
		elapsed_text = elapsed_text.ljust(self._max_lengths['elapsed'])
		return colour(text=elapsed_text, text_colour=self._full_colour)

	@property
	def _remaining_time_string(self):
		remaining_text = f'r:{self.format_time(self.remaining_seconds)} '
		self._max_lengths['remaining'] = max(self._max_lengths['remaining'], len(remaining_text))
		remaining_text = remaining_text.ljust(self._max_lengths['remaining'])
		return colour(text=remaining_text, text_colour=self._empty_colour)

	@property
	def _percent_string(self):
		percent_text = f' {self.percent_formatted}'
		self._max_lengths['percent'] = max(self._max_lengths['percent'], len(percent_text))
		percent_text = percent_text.rjust(self._max_lengths['percent'])
		return colour(text=percent_text, text_colour=self._full_colour)

	def _get_text_string(self, text):
		text_text = f' {text}'
		self._max_lengths['text'] = max(self._max_lengths['text'], len(text_text))
		text_text = text_text.ljust(self._max_lengths['text'])
		return colour(text=text_text, text_colour=self._text_colour)

	@property
	def _main_bar(self):
		if self._parent:
			return self._parent
		else:
			return self

	def show(self, amount, total=None, extra_amount=0, percent=True, bar=True, time=True, text='', next_line=None):
		"""
		:type amount: int or float or NoneType
		:type total: NoneType or float
		:type extra_amount: int or float or NoneType
		:type percent: bool
		:type bar: bool
		:type time: bool
		:type text: str
		"""
		if total is not None:
			self.set_total(total=total)

		if amount is not None:
			self.amount = amount + extra_amount
		else:
			self.amount = amount

		if self._echo:
			time_now = datetime.now()
			if self._shown < 1 or elapsed(start_time=self._display_time, end_time=time_now) > self._display_wait or self.amount >= self.total:
				self._shown += 1
				self._display_time = time_now

				string = ''
				try:
					string += self._animation_string

					if time:
						string += self._main_bar._elapsed_time_string
						string += self._main_bar._remaining_time_string

					if bar:
						self._last_formatted_bar = self.bar
						string += self._last_formatted_bar

					if percent:
						string += self._main_bar._percent_string

					if self._parent is None:
						self._last_text = text
					else:
						if len(self._parent._last_text) > 0:
							self._last_text = self._parent._last_text + ' / ' + text
						else:
							self._last_text = text

					string += self._get_text_string(text=self._last_text)

					self._max_lengths['with_colour'] = max(self._max_lengths['with_colour'], len(string))
					string = string.ljust(self._max_lengths['with_colour'])

				except Exception as e:
					self.write(string=f'progress bar error: {e}')
					raise e

				if self.amount >= self.total:
					self.complete()

				if self._echo:
					if self._completed:
						if self._disappear:
							self.write(string=' ' * self._max_lengths['with_colour'] + ' ')

						else:
							self.write(string=string)

						if next_line is None:
							next_line = self._next_line

						if next_line:
							print('')
					else:
						self.write(string=string, flush=True)

	@classmethod
	def map(
			cls, function, iterable, percent=True, bar=True, time=True, text='', echo=1,
			next_line=True, iterable_text=None,
			**kwargs
	):
		echo = max(0, echo)

		def _func(x, _progress, _progress_bar, _text=''):
			if progress['amount'] == 0:
				progress_bar.show(
					amount=progress['amount'], percent=percent, bar=bar, time=time, text=text + _text
				)

			function_result = function(x)

			_progress['amount'] += 1
			if elapsed(start_time=progress['update_time'], end_time=datetime.now()) > 0.5 or _progress['amount'] >= total:
				_progress_bar.show(amount=_progress['amount'], percent=percent, bar=bar, time=time, text=text + _text)
				_progress['update_time'] = datetime.now()

			return function_result

		total = len(iterable)

		progress = {'amount': 0, 'update_time': datetime.now()}
		progress_bar = cls(total=total, next_line=next_line, **kwargs)

		if echo:
			if iterable_text is None:
				result = _map(
					lambda x: _func(x=x, _progress=progress, _progress_bar=progress_bar),
					iterable
				)
			else:
				result = _map(
					lambda x: _func(x=x[0], _text=x[1], _progress=progress, _progress_bar=progress_bar),
					zip(iterable, iterable_text)
				)

			return result
		else:
			return _map(function, iterable)

	@classmethod
	def apply(
			cls, function, data=None, series=None, percent=True, bar=True, time=True, text='',
			axis=1, echo=1, next_line=True, **kwargs
	):
		"""
		:type function: function
		:type data: pd.DataFrame
		:type series: pd.Series
		:type percent: bool
		:type bar: bool
		:type time: bool
		:type text: str
		:type axis: int
		:type echo: bool or int
		:type next_line: bool
		:rtype: pd.DataFrame or pd.Series
		"""
		echo = max(0, echo)
		# either data or series should be provided:
		if data is None and series is None:
			raise ValueError('either data or series should be provided')

		if data is not None and series is not None:
			raise ValueError('both data and series cannot be provided')

		def _func(x, _progress, _progress_bar):
			function_result = function(x)
			_progress['amount'] += 1

			if elapsed(start_time=_progress['update_time'], end_time=datetime.now()) > 0.5 or _progress['amount'] >= total:
				_progress_bar.show(amount=_progress['amount'], percent=percent, bar=bar, time=time, text=text)
				_progress['update_time'] = datetime.now()

			return function_result

		if data is not None:
			if axis == 1:
				total = data.shape[0]
			else:
				total = data.shape[1]
		else:
			total = series.shape[0]

		if total == 0:
			return None

		progress = {'amount': 0, 'update_time': datetime.now()}
		progress_bar = cls(total=total, next_line=next_line, **kwargs)
		if echo:
			progress_bar.show(amount=0, percent=percent, bar=bar, time=time, text=text)
			if data is not None:
				result = data.apply(func=lambda x: _func(x=x, _progress=progress, _progress_bar=progress_bar), axis=axis)
			else:
				result = series.apply(func=lambda x: _func(x=x, _progress=progress, _progress_bar=progress_bar))

		else:
			if data is not None:
				result = data.apply(func=function, axis=axis)
			else:
				result = series.apply(func=function)

		return result

	def __gt__(self, other):
		if isinstance(other, self.__class__):
			return self._echo > other._echo
		else:
			try:
				return self._echo > other
			except TypeError:
				return True

	def __lt__(self, other):
		if isinstance(other, self.__class__):
			return self._echo < other._echo
		else:
			try:
				return self._echo < other
			except TypeError:
				return False

	def __ge__(self, other):
		return not (self < other)

	def __le__(self, other):
		return not (self > other)

	def __eq__(self, other):
		if isinstance(other, self.__class__):
			return self._echo == other._echo
		else:
			return False

	def __ne__(self, other):
		return not (self == other)

	@classmethod
	def test(cls, time=1, n=10000, bar_characters=None):
		bar = cls(total=n, display_wait=0.1, bar_characters=bar_characters)
		for i in range(n):
			bar.show(amount=i, text='testing ...')
			sleep(time / n / 2)
		bar.show(amount=n, text='test done!')



from pandas import DataFrame
import random


def split_by_id(data, id_column, n=None, fraction=None, random_state=None, **kwargs):
	"""
	this method splits the data, by considering rows with the same id as one and refusing to divide them
	it is useful when you want a sample of the data based on ids when the ids are repeated
	and you want to take all or none of each id, but not a partial sample of multiple rows with the same id
	:param DataFrame data: dataframe to be sampled
	:param str id_column: column to be used as unique identifier
	:param int n: number of items to return
	:param float fraction: fraction of items to return, cannot be used with `n`
	:param bool replace: with or without replacement
	:param int random_state: seed for the random number generator
	:rtype: tuple[DataFrame, DataFrame]
	"""
	# find unique ids:
	ids = list(set(data[id_column]))

	# shuffle the ids:
	random.Random(random_state).shuffle(ids)

	# choose training and test ids
	if n is None:
		n = max(0, min(len(ids), round(fraction * len(ids))))
	training_ids = ids[:n]
	test_ids = ids[n:]

	data = data.copy()
	data['__index__'] = data.index
	data['__order__'] = range(len(data))

	# create training and test dataframes
	training_data = DataFrame({id_column: training_ids})
	test_data = DataFrame({id_column: test_ids})

	# merge with the rest of the data
	training_data = training_data.merge(
		right=data, on=id_column, how='left'
	).set_index('__index__').sort_values('__order__').drop(columns='__order__')
	training_data.index.name = data.index.name

	test_data = test_data.merge(
		right=data, on=id_column, how='left'
	).set_index('__index__').sort_values('__order__').drop(columns='__order__')
	test_data.index.name = data.index.name

	return training_data, test_data


def sample_by_id(data, id_column, n=None, fraction=None, replace=False, random_state=None, **kwargs):
	"""
	this method samples the data, by considering rows with the same id as one and refusing to divide them
	it is useful when you want a sample of the data based on ids when the ids are repeated
	and you want to take all or none of each id, but not a partial sample of multiple rows with the same id
	:param DataFrame data: dataframe to be sampled
	:param str id_column: column to be used as unique identifier
	:param int n: number of items to return
	:param float fraction: fraction of items to return, cannot be used with `n`
	:param bool replace: with or without replacement
	:param int random_state: seed for the random number generator
	:rtype: DataFrame
	"""
	data = data.copy()
	data['__index__'] = data.index
	ids = data[[id_column]].drop_duplicates()
	sampled_ids = ids.sample(n=n, frac=fraction, replace=replace, random_state=random_state, **kwargs)
	result = sampled_ids.merge(right=data, on=id_column, how='left').set_index('__index__')
	result.index.name = data.index.name
	return result



from .Colour import Colour


class Gradient:
	def __init__(self, colour_1, colour_2, num_levels=10):
		colour_1 = Colour(colour_1)
		colour_2 = Colour(colour_2)

		self._colours = [
			(colour_1 * (num_levels - i - 1)).mix(colour_2 * i)
			for i in range(num_levels)
		]
		self._num_levels = num_levels

	def __repr__(self):
		return '\n'.join([repr(colour) for colour in self._colours])

	def colourize(self, string):
		"""
		:type string: str
		:rtype: str
		"""
		denominator = len(string) - 1
		return ''.join([
			self.get(ratio=numerator / denominator).colourize(string=character, background=None)
			for numerator, character in enumerate(string)
		])

	def get(self, ratio):
		"""
		:type ratio: float
		:return: Colour
		"""
		ratio = max(0.0, min(1.0, ratio))
		return self._colours[round(ratio * (self._num_levels - 1))]



from ._align import align
from ._wrap import wrap
from ._convert_camel_to_snake import convert_camel_to_snake
from ._remove_non_alphanumeric import remove_non_alphanumeric
from ._remove_non_alphanumeric import remove_non_alphabetic
from ._is_word import is_word
from ._is_word import count_words
from ._count_sentences import count_sentences
from ._split_and_strip import split_and_strip



from datetime import date


def get_number_of_days(year, month=None):
	"""
	returns the number of days in a month or a year
	:param int year: the year at question
	:param int or NoneType month: if None, the function returns the number of days in a year
	:rtype: int
	"""
	if month:
		if month == 12:
			return (date(year + 1, 1, 1) - date(year, 12, 1)).days
		else:
			return (date(year, month + 1, 1) - date(year, month, 1)).days
	else:
		return (date(year + 1, 1, 1) - date(year, 1, 1)).days


from ...exceptions import FunctionNotImplementedError
from pandas import DataFrame


def get_display_function():
	try:
		from IPython.core.display import display
		return display
	except ImportError:
		return False


class DataContainer:
	def __init__(self, data, x_columns=None, y_column=None, sort_columns=None):
		self._data = data
		self._x_columns = x_columns
		self._y_column = y_column
		self._sort_columns = sort_columns

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return self._data

	@property
	def columns(self):
		return self._data.columns

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text(repr(self))
		else:
			self.display(p=p)

	def display(self):
		raise FunctionNotImplementedError('display function is not implemented for DataContainer!')



def select_columns(data, include = None, exclude = None):
	if include is None:
		include = list(data.columns)

	if exclude is None:
		exclude = []

	columns = [column for column in include if column not in exclude]
	return data[columns]



from joblib import Parallel, delayed
from .BaseController import BaseController
from ._WorkerReport import WorkerReport
from ._Task import Task, Outcome
from ..time.progress import iterate
from ._DEFAULT_VALUES import TIME_UNIT, N_JOBS


class JobController(BaseController):
	def __init__(self, time_unit=TIME_UNIT, n_jobs=N_JOBS):
		super().__init__(time_unit=time_unit)
		self._n_jobs = n_jobs

	def do(self, echo=1):
		worker_id = self._generate_worker_id()
		report = WorkerReport(worker_id=worker_id)

		processor = Parallel(n_jobs=self._n_jobs, backend='threading', require='sharedmem')

		def _do_task(task):
			"""
			:type task: Task
			:rtype: Outcome
			"""
			return task.do(worker_id=worker_id)

		tasks = []
		while self.to_do_count > 0:
			task = self.to_do_queue.popleft()
			self._tasks_status[task.id] = 'started'
			tasks.append(task)

		outcomes = processor(
			delayed(_do_task)(task=task)
			for task in iterate(tasks, echo=echo)
		)

		for outcome in outcomes:
			self.done_queue.append(outcome)
			self._tasks_status[outcome.task_id] = 'done'
			report.add_task_id(task_id=outcome.task_id)

		report.end()
		if echo:
			self.write(f'To-do: {self.to_do_count} - Doing: 0 - Done: {self.done_count} - Processed: 0       ')
		self.process_done_queue(echo=echo)



def cross_two_lists(l1, l2):
	"""
	:rtype: list[tuple]
	"""
	return [(x, y) for x in l1 for y in l2]


def cross_lists(*args):
	"""
	:rtype: list[tuple]
	"""
	if len(args) == 2:
		return cross_two_lists(args[0], args[1])
	elif len(args) > 2:
		result = cross_lists(cross_two_lists(args[0], args[1]), *args[2:])
		return [(x[0][0], x[0][1], *x[1:]) for x in result]
	elif len(args) == 1 and isinstance(args[0], (list, tuple)):
		return cross_lists(*args[0])



from pyspark.sql.functions import array, col, explode, lit
from pyspark.sql.functions import create_map
from pyspark.sql import DataFrame as SparkDF
from pandas import DataFrame as PandasDF
from typing import Iterable
from itertools import chain


def melt(data, id_columns, value_columns, variable_name='variable', value_name='value'):
	"""
	converts a DataFrame from wide to long format
	:type data: SparkDF or PandasDF
	:type id_columns: str or list[str]
	:type value_columns: str or list[str]
	:type variable_name: str
	:type value_name: str
	:rtype: SparkDF or PandasDF
	"""



from pandas import DataFrame, Series

def set_value_where(data, columns, value, where):
	"""
	:type data: DataFrame
	:type columns: str
	:type where: Series
	:rtype: DataFrame
	"""

	data.set_value(index=data[where].index, col=columns, value=value)
	return data



from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeClassifier
from pandas.api.types import is_numeric_dtype
from copy import deepcopy
from ...time.progress import iterate
from ...multiprocessing import SimpleController
from ...collections import get_intersection
from ._ColumnImputer import ColumnImputer


class DataImputer:
	def __init__(
			self, estimators, ignore_columns=None, helper_columns=None, impute_columns=None,
			cross_validation=None, drop_na=False, main_metric='rmse', impute_categorical=False,
			controller=None
	):
		"""
		:type estimators: LinearRegression or DecisionTreeClassifier or list
		:type ignore_columns: NoneType or str or list[str]
		:type helper_columns: NoneType or str or list[str]
		:type impute_columns: NoneType or str or list[str]
		:type cross_validation: CrossValidation or NoneType
		:type drop_na: bool
		:type main_metric: str
		:type impute_categorical: bool
		:type controller: BaseController
		"""
		if ignore_columns is None:
			ignore_columns = []
		elif isinstance(ignore_columns, str):
			ignore_columns = [ignore_columns]
		elif not isinstance(ignore_columns, list):
			raise TypeError(f'ignore_columns cannot be of type {type(ignore_columns)}')

		self._ignore_columns = ignore_columns

		if helper_columns is not None:
			if isinstance(helper_columns, str):
				helper_columns = [helper_columns]

			if len(get_intersection(helper_columns, ignore_columns)) > 0:
				raise ValueError(f'helper_columns and ignore_columns intersect: {get_intersection(helper_columns, ignore_columns)}')

		self._helper_columns = helper_columns

		if impute_columns is not None:
			if isinstance(impute_columns, str):
				impute_columns = [impute_columns]

			if len(get_intersection(impute_columns, ignore_columns)) > 0:
				raise ValueError(f'impute_columns and ignorre_columns intersect: {get_intersection(impute_columns, ignore_columns)}')

		self._impute_columns = impute_columns

		if not isinstance(estimators, (list, tuple)):
			estimators = [estimators]

		self._estimators = estimators
		self._cross_validation = cross_validation
		self._drop_na = drop_na
		self._main_metric = main_metric.lower()
		self._impute_categorical = impute_categorical
		self._column_imputers = {}
		self._controller = controller

	def _get_ignore_columns(self, data):
		return [col for col in data.columns if col in self._ignore_columns]

	def _get_helper_columns(self, data, ignore_columns):
		if self._helper_columns is not None:
			return self._helper_columns
		else:
			return [col for col in data.columns if col not in ignore_columns]

	def _get_impute_columns(self, data, ignore_columns):
		if self._impute_columns is not None:
			return self._impute_columns
		else:
			impute_columns = [
				col for col in data.columns
				if col not in ignore_columns and data[col].isnull().values.any()
			]
			if self._impute_categorical:
				return impute_columns
			else:
				return [col for col in impute_columns if is_numeric_dtype(data[col])]

	def _get_controller(self):
		"""
		:rtype: SimpleController or NoneType
		"""
		return self._controller

	def _get_estimators(self):
		return [deepcopy(estimator) for estimator in self._estimators]

	def fit(self, X, echo=1):
		ignore_columns = self._get_ignore_columns(data=X)
		helper_columns = self._get_helper_columns(data=X, ignore_columns=ignore_columns)
		impute_columns = self._get_impute_columns(data=X, ignore_columns=ignore_columns)

		def _create_column_imputer(column):
			column_imputer = ColumnImputer(
				column=column, estimators=self._get_estimators(), helper_columns=helper_columns,
				cross_validation=self._cross_validation, drop_na=self._drop_na,
				main_metric=self._main_metric
			)
			column_imputer.fit(X=X)
			return column_imputer

		if self._controller is None:
			self._column_imputers = {
				column: _create_column_imputer(column=column)
				for column in iterate(impute_columns)
			}

		else:
			controller = self._get_controller()
			for column in impute_columns:
				controller.add_task(
					function=_create_column_imputer, kwargs={'column': column},
					task_id=f'{column}_imputer'
				)
			controller.do(echo=echo)
			self._column_imputers = {
				column: controller.get_result(task_id=f'{column}_imputer')
				for column in impute_columns
			}

	def transform(self, X, echo=1):
		X = X.copy()

		def _impute_column(imputer):
			"""
			:type imputer: ColumnImputer
			:rtype: list
			"""
			return {'name': imputer._imputed_column, 'values': imputer._transform(X=X, return_y=True)}

		if self._controller is None:
			column_names_and_values = [
				_impute_column(imputer=imputer)
				for imputer in iterate(self._column_imputers.values())
			]

		else:
			controller = self._get_controller()
			for imputer in self._column_imputers.values():
				controller.add_task(
					function=_impute_column, kwargs={'imputer': imputer},
					task_id=f'{imputer._imputed_column}_values'
				)
			controller.do(echo=echo)
			column_names_and_values = [
				controller.get_result(task_id=f'{column}_values')
				for column in self._column_imputers.keys()
			]

		for d in column_names_and_values:
			name = d['name']
			values = d['values']
			X[name] = values

		return X



from pandas import DataFrame
from ._get_cross_validation import get_cross_validation
from ._get_training_test import get_training_test
from ._ValidationContainer import ValidationContainer
from ._TrainingTestContainer import TrainingTestContainer
from ...exceptions import MissingArgumentError


class Validation:
	def __init__(
			self, num_splits, id_columns=None, sort_columns=None, random_state=None,
			holdout_ratio=None, holdout_count=None, test_ratio=None, test_count=None,
			min_training_count=None, min_training_ratio=None
	):
		"""
		:type id_columns: NoneType or list[str] or str
		:type sort_columns: NoneType or list[str] or str
		:type random_state: NoneType or int or float
		"""
		if isinstance(id_columns, str):
			id_columns = [id_columns]
		if isinstance(sort_columns, str):
			sort_columns = [sort_columns]

		self._num_splits = num_splits
		self._id_columns = id_columns
		self._sort_columns = sort_columns
		self._random_state = random_state

		self._holdout_ratio = holdout_ratio
		self._holdout_count = holdout_count
		self._test_ratio = test_ratio
		self._test_count = test_count
		self._min_training_count = min_training_count
		self._min_training_ratio = min_training_ratio

	def split(self, data, random_state=None):
		"""
		:type data: DataFrame
		:type random_state: int
		:rtype: ValidationContainer
		"""
		return get_cross_validation(
			data=data,
			num_splits=self._num_splits,
			holdout_ratio=self._holdout_ratio,
			holdout_count=self._holdout_count,
			test_count=self._test_count,
			sort_columns=self._sort_columns,
			id_columns=self._id_columns,
			test_ratio=self._test_ratio,
			random_state=random_state or self._random_state,
			min_training_count=self._min_training_count,
			min_training_ratio=self._min_training_ratio
		)


class TrainingTest(Validation):
	def __init__(self, id_columns=None, sort_columns=None, random_state=None, test_ratio=None, test_count=None):
		"""
		:param id_columns:
		:param sort_columns:
		:param random_state:
		:param test_ratio:
		:param test_count:
		"""
		super().__init__(
			num_splits=None, id_columns=id_columns, sort_columns=sort_columns, random_state=random_state,
			holdout_ratio=None, holdout_count=None, test_count=test_count, test_ratio=test_ratio,
			min_training_count=None, min_training_ratio=None
		)

	def split(self, data, random_state=None):
		"""
		:type data: DataFrame
		:type random_state: int
		:rtype: TrainingTestContainer
		"""
		return get_training_test(
			data=data,
			sort_columns=self._sort_columns,
			id_columns=self._id_columns,
			test_ratio=self._test_ratio,
			test_count=self._test_count,
			random_state=random_state or self._random_state
		)


class TimeSeriesTrainingTest(TrainingTest):
	def __init__(self, sort_columns, random_state=None, test_ratio=None, test_count=None):
		"""
		:param id_columns:
		:param sort_columns:
		:param random_state:
		:param test_ratio:
		:param test_count:
		"""
		super().__init__(
			id_columns=None, sort_columns=sort_columns, random_state=random_state,
			test_ratio=test_ratio, test_count=test_count
		)


class CrossValidation(Validation):
	def __init__(
			self, num_splits, id_columns=None, random_state=None,
			holdout_ratio=None, holdout_count=None, test_ratio=None, test_count=None,
			min_training_count=None, min_training_ratio=None
	):
		"""
		:param num_splits:
		:param id_columns:
		:param random_state:
		:param holdout_ratio:
		:param holdout_count:
		:param test_ratio:
		:param test_count:
		:param min_training_count:
		:param min_training_ratio:
		"""
		super().__init__(
			num_splits=num_splits,
			id_columns=id_columns, sort_columns=None, random_state=random_state,
			holdout_ratio=holdout_ratio, holdout_count=holdout_count, test_ratio=test_ratio,
			test_count=test_count,
			min_training_count=min_training_count, min_training_ratio=min_training_ratio
		)


class TimeSeriesValidation(Validation):
	def __init__(
			self, num_splits, sort_columns, random_state=None,
			holdout_ratio=None, holdout_count=None, test_ratio=None, test_count=None,
			min_training_count=None, min_training_ratio=None
	):
		"""
		:param num_splits:
		:param sort_columns:
		:param random_state:
		:param holdout_ratio:
		:param holdout_count:
		:param test_ratio:
		:param test_count:
		:param min_training_count:
		:param min_training_ratio:
		"""
		if sort_columns is None:
			raise MissingArgumentError('sort_columns should be provided!')

		super().__init__(
			num_splits=num_splits,
			id_columns=None, random_state=random_state, sort_columns=sort_columns,
			holdout_ratio=holdout_ratio, holdout_count=holdout_count, test_ratio=test_ratio,
			test_count=test_count,
			min_training_count=min_training_count, min_training_ratio=min_training_ratio
		)



from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame as SparkDF
from ...collections import get_duplicates

def assert_columns_are_unique(df):
    """
    :type df: PandasDF or SparkDF
    """
    df_columns = list(df.columns)
    duplicates = get_duplicates(df_columns)

    if len(duplicates) > 0:
        message = f'Duplicate columns: "{", ".join(duplicates)}"'
        print(message)
        raise KeyError(message)
    else:
        return True



def fill_na(data, replace_with=0, cols=None, except_cols=None, na_col_suffix='_na', inplace=False, na_col_as_int=False):
	except_cols = except_cols or []

	if inplace:
		new_data = data
	else:
		new_data = data.copy()

	if cols is None:
		cols = new_data.columns

	for col in cols:
		if col not in except_cols and new_data[col].isnull().values.any():
			na_values = new_data[col].isnull().values
			if na_col_as_int:
				na_values = na_values.astype(int)  # to turn boolean to int
			new_data[str(col) + na_col_suffix] = na_values
			new_data[col].fillna(value=replace_with, inplace=True)

	return new_data



import pandas as pd
from ...text import convert_camel_to_snake
from ._standardize_columns import standardize_columns as standardize


def read_excel(
		path, standardize_columns=True, headers=0, column_names=None, index_columns=None,
		find_header_by_columns=None, ignore_sheets_with_no_header=False
):
	"""
	:param str 						path: 			the path to the excel file
	:param bool 		standardize_columns: 		if True the columns will be converted to snake_case
	:param int or dict[str, int] 	headers: 		header index, if dictionary then each sheet will use the respective
													one based on the key and the sheet name
	:param list or dict[str, list] 	column_names: 	explicit column_names, also dictionary is used to point to each sheet
	:param int or dict[str, int] 	index_columns: 	tells which column should be used as index, can also be dictionary

	:param dict[str, list] 	find_header_by_columns: if the header is not obvious, this allows the function to find it
	:param bool		ignore_sheets_with_no_headers: 	relevant when find_header_by_columns, if True,
													when header is not found, sheet is ignored
	"""
	excel = pd.ExcelFile(path)
	result = {}
	for sheet_name in excel.sheet_names:
		snake_name = convert_camel_to_snake(sheet_name)

		if isinstance(headers, dict):
			if sheet_name in headers:
				_header = headers[sheet_name]
			elif snake_name in headers:
				_header = headers[snake_name]
			else:
				raise KeyError(f'neither {sheet_name} nor {snake_name} in header dictionary!')
		else:
			_header = headers

		if isinstance(column_names, dict):
			if sheet_name in column_names:
				_column_names = column_names[sheet_name]
			elif snake_name in column_names:
				_column_names = column_names[snake_name]
			else:
				raise KeyError(f'neither {sheet_name} nor {snake_name} in column_names dictionary!')
		else:
			_column_names = column_names

		if isinstance(index_columns, dict):
			if sheet_name in index_columns:
				_index_column = index_columns[sheet_name]
			elif snake_name in index_columns:
				_index_column = index_columns[snake_name]
			else:
				raise KeyError(f'neither {sheet_name} nor {snake_name} in index_columns dictionary!')
		else:
			_index_column = index_columns

		if isinstance(find_header_by_columns, dict):
			if sheet_name in find_header_by_columns:
				_find_header_by_columns = find_header_by_columns[sheet_name]
			elif snake_name in find_header_by_columns:
				_find_header_by_columns = find_header_by_columns[snake_name]
			else:
				raise KeyError(f'neither {sheet_name} nor {snake_name} in find_header_by_columns dictionary!')
		else:
			_find_header_by_columns = find_header_by_columns

		if _find_header_by_columns is None:
			result[snake_name] = pd.read_excel(
				path, sheet_name=sheet_name, header=_header, names=_column_names, index_col=_index_column
			)
		else:
			temp = pd.read_excel(
				path, sheet_name=sheet_name, header=None, names=_column_names, index_col=_index_column
			)
			if not isinstance(_find_header_by_columns, list):
				_find_header_by_columns = [_find_header_by_columns]
			for i, row in temp.iterrows():
				if all(
						convert_camel_to_snake(x, ignore_errors=True) == convert_camel_to_snake(y, ignore_errors=True)
						for x, y in zip(_find_header_by_columns, row)
				):
					result[snake_name] = pd.read_excel(
						path, sheet_name=sheet_name, header=i, names=_column_names, index_col=_index_column
					)
					break
			else:
				if ignore_sheets_with_no_header:
					continue
				else:
					raise KeyError(f'could not find the header using columns: {_find_header_by_columns}')

	if standardize_columns:
		result = {key: standardize(value) for key, value in result.items()}

	return result



from ..evaluation import evaluate_regression, evaluate_classification


def train_and_test(
		estimator, evaluation_type, training_data, test_data, x_columns, y_column, sort_columns,
		ignore_missing_y=False, ignore_missing_x=False, replace_missing_x='mean'
):
	training_data = training_data[x_columns + [y_column]]

	if ignore_missing_y and ignore_missing_y:
		training_data = training_data.drop_na()
		test_data = test_data.drop_na()

	elif ignore_missing_y:
		training_data = training_data[training_data[y_column].notna()]
		test_data = test_data[training_data[y_column].notna()]

	elif ignore_missing_x:
		training_data = training_data[~training_data[x_columns].isnull().any(axis=1)]
		test_data = test_data[~test_data[x_columns].isnull().any(axis=1)]

	if not ignore_missing_x:
		if replace_missing_x == 'mean':
			training_data[x_columns] = training_data[x_columns].fillna(training_data[x_columns].mean())
			test_data[x_columns] = test_data[x_columns].fillna(test_data[x_columns].mean())

		elif replace_missing_x == 'median':
			training_data[x_columns] = training_data[x_columns].fillna(training_data[x_columns].median())
			test_data[x_columns] = test_data[x_columns].fillna(test_data[x_columns].median())

	estimator.fit(X=training_data[x_columns], y=training_data[y_column])
	actual = test_data[y_column]
	predicted = estimator.predict(test_data[x_columns])
	if evaluation_type.lower().startswith('regress'):
		evaluation = evaluate_regression(actual=actual, predicted=predicted)
	elif evaluation_type.lower().startswith('class'):
		evaluation = evaluate_classification(actual=actual, predicted=predicted)

	evaluation = {
		'training_size': training_data.shape[0],
		'test_size': test_data.shape[0],
		**evaluation
	}

	if sort_columns is not None:
		sort_values = {}
		for sort_column in sort_columns:
			sort_values[f'training_from_{sort_column}'] = training_data[sort_column].min()
			sort_values[f'training_to_{sort_column}'] = training_data[sort_column].max()
			sort_values[f'test_from_{sort_column}'] = test_data[sort_column].min()
			sort_values[f'test_to_{sort_column}'] = test_data[sort_column].max()
		evaluation = {**sort_values, **evaluation}
	return evaluation



from setuptools import setup, find_packages


def readme():
	with open('./README.md') as f:
		return f.read()


setup(
	name='atlantis',
	version='2021.11.19',
	description='Python library for simplifying data science',
	long_description=readme(),
	long_description_content_type='text/markdown',
	url='https://github.com/idin/atlantis',
	author='Idin',
	author_email='py@idin.ca',
	license='MIT',
	packages=find_packages(exclude=("jupyter", ".idea", ".git", "data_files")),
	install_requires=[
		'base32hex', 'geopy', 'pandas', 'joblib', 'numpy', 'sklearn', 'multiprocess', 'pyspark', 'matplotlib',
		'scipy'
	],
	package_data={'atlantis': ['data_files/*.pickle']},
	python_requires='~=3.6',
	zip_safe=False
)



import numpy as np
from pandas import DataFrame


def count_rows_with_missing(data):
	"""
	:type data: DataFrame
	:rtype: int
	"""
	return np.count_nonzero(np.count_nonzero(data.isnull(), axis=1))


def count_rows_without_missing(data):
	"""
	:type data: DataFrame
	:rtype: int
	"""
	return data.shape[0] - count_rows_with_missing(data=data)



from sklearn.linear_model import LinearRegression, LogisticRegression
from ...collections import create_grid
from ...hash import hash_object


class EstimatorGrid:
	def __init__(self, estimator, estimator_arguments=None):
		"""
		:type estimator: LogisticRegression or LinearRegression
		:type estimator_arguments: dict
		"""
		self._estimator = estimator
		self._hash_to_id = {}
		self._id_to_arguments = {}

		if estimator_arguments is not None:
			self.append(estimator_arguments=estimator_arguments)

	def get_available_id(self):
		if len(self._id_to_arguments) == 0:
			return 1
		else:
			max_id = max([i for i in self._id_to_arguments.keys()])

			if len(self._id_to_arguments) == max_id:
				return max_id + 1

			else:
				for i in range(1, max_id + 2):
					if i not in self._id_to_arguments:
						return i

	@property
	def estimator_ids(self):
		"""
		:rtype: list[int]
		"""
		return list(self._id_to_arguments.keys())

	def __contains__(self, item):
		return item in self._id_to_arguments

	@property
	def estimator(self):
		"""
		:rtype: type
		"""
		return self._estimator

	@property
	def name(self):
		return self.estimator.__name__

	def append(self, estimator_arguments):
		"""
		adds a dictionary of arguments if they don't exist
		:type estimator_arguments: dict
		"""
		grid = create_grid(dictionary=estimator_arguments)

		for dictionary in grid:
			hash_key = hash_object(dictionary)

			if hash_key not in self._hash_to_id:
				new_id = self.get_available_id()
				self._hash_to_id[hash_key] = new_id
				self._id_to_arguments[new_id] = dictionary

	@property
	def estimator_dictionaries(self):
		"""
		:rtype: list[dict]
		"""
		return [
			{'estimator': self.estimator, 'estimator_arguments': estimator_arguments, 'id': f'{self.name}_{key}'}
			for key, estimator_arguments in self._id_to_arguments.items()
		]

	def __add__(self, other):
		"""
		:type other: EstimatorGrid
		:rtype: EstimatorGrid
		"""
		result = EstimatorGrid(estimator=self.estimator)

		for estimator_arguments in self._id_to_arguments.values():
			result.append(estimator_arguments=estimator_arguments)

		for estimator_arguments in other._id_to_arguments.values():
			result.append(estimator_arguments=estimator_arguments)

		return result


class EstimatorRepository:
	def __init__(self, estimator=None, estimator_arguments=None):
		self._estimator_grids_dictionary = {}
		if estimator is not None and estimator_arguments is not None:
			self.append(estimator=estimator, estimator_arguments=estimator_arguments)

	def __contains__(self, item):
		"""
		:type item: tuple or str
		"""
		if isinstance(item, str):
			return item in self.estimator_grids
		else:
			estimator_name, estimator_id = item
			if estimator_name not in self.estimator_grids:
				return False
			else:
				return estimator_id in self.estimator_grids[estimator_name]

	def append(self, estimator, estimator_arguments=None):
		"""
		:type estimator: type
		:type estimator_arguments: dict
		"""
		estimator_arguments = estimator_arguments or {}
		class_name = estimator.__name__

		if class_name not in self._estimator_grids_dictionary:
			self._estimator_grids_dictionary[class_name] = EstimatorGrid(estimator=estimator)

		self._estimator_grids_dictionary[class_name].append(estimator_arguments=estimator_arguments)

	@property
	def estimator_grids(self):
		"""
		:rtype: list[EstimatorGrid]
		"""
		return list(self._estimator_grids_dictionary.values())

	@property
	def estimator_dictionaries(self):
		"""
		:rtype: list[dict]
		"""
		return [
			dictionary
			for grid in self.estimator_grids
			for dictionary in grid.estimator_dictionaries
		]

	def __add__(self, other):
		"""
		:type other: EstimatorRepository
		"""
		result = EstimatorRepository()

		for database in self.estimator_grids:
			result.estimator_grids[database.name] = database

		for database in other.estimator_grids:
			if database.name in result.estimator_grids:
				result.estimator_grids[database.name] += database
			else:
				result.estimator_grids[database.name] = database

		return result



from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame as SparkDF
from pyspark.sql import functions as f
from pyspark.sql.types import FloatType, BooleanType

PERCENTILE75 = 0.6745


def get_modified_zscore(data, column, zscore_column=None, relative_error=1e-6):
	"""

	:type data: PandasDF or SparkDF
	:type column: str
	:type zscore_column: str or NoneType
	:type relative_error: float
	:rtype PandasDF or SparkDF
	"""
	if zscore_column is None:
		zscore_column = f'{column}_zscore'
	if isinstance(data, PandasDF):
		data = data.copy()
		median = data[column].median()
		deviation = data[column] - median
		mad = deviation.abs().median()
		data[zscore_column] = PERCENTILE75 * deviation / mad

	elif isinstance(data, SparkDF):
		median = data.approxQuantile(column, [0.5], relative_error)[0]
		data = data.withColumn(
			'_deviation_', f.col(column) - f.lit(median)
		).withColumn('_abs_deviation_', f.abs('_deviation_'))
		mad = data.approxQuantile('_abs_deviation_', [0.5], relative_error)[0]

		@f.udf(FloatType())
		def _get_modified_zscore(deviation):
			return PERCENTILE75 * deviation / mad

		data = data.withColumn(zscore_column, _get_modified_zscore('_deviation_')).drop(
			'_abs_deviation_', '_deviation_'
		)

	else:
		raise TypeError(f'data of type {type(data)} is not supported!')

	return data


def is_outlier(data, column, is_outlier_column=None, threshold=3.5, relative_error=1e-6):
	"""

	:type data: PandasDF or SparkDF
	:type column: str
	:type is_outlier_column: str or NoneType
	:type threshold: float
	:type relative_error: float
	:rtype PandasDF or SparkDF
	"""
	if is_outlier_column is None:
		is_outlier_column = f'{column}_is_outlier'

	data = get_modified_zscore(data=data, column=column, zscore_column='_zscore_', relative_error=relative_error)
	if isinstance(data, PandasDF):
		data[is_outlier_column] = data['_zscore_'].abs() > threshold
		data = data.drop(columns=['_zscore_'])

	elif isinstance(data, SparkDF):

		@f.udf(BooleanType())
		def _is_outlier(x):
			return abs(x) > threshold

		data = data.withColumn(is_outlier_column, _is_outlier('_zscore_')).drop('_zscore_')

	else:
		raise TypeError(f'data of type {type(data)} is not supported!')

	return data


def remove_outliers(data, column, threshold=3.5, relative_error=1e-6):
	"""
	:type data: PandasDF or SparkDF
	:type column: str
	:type threshold: float
	:type relative_error: float
	:rtype PandasDF or SparkDF
	"""
	data = is_outlier(
		data=data, column=column, is_outlier_column='_is_outlier_', threshold=threshold, relative_error=relative_error
	)

	if isinstance(data, PandasDF):
		return data[~data['_is_outlier_']].drop(columns=['_is_outlier_'])
	elif isinstance(data, SparkDF):
		return data.filter(~f.col('_is_outlier_')).drop('_is_outlier_')
	else:
		raise TypeError(f'data of type {type(data)} is not supported!')



from pyspark.sql import DataFrame as SparkDF
from pandas import DataFrame as PandasDF, Series as PandasSeries
from sklearn.linear_model import LogisticRegression as SkleanLR
from pyspark.ml.classification import LogisticRegression as SparkLR


class LogisticRegression:
	def __init__(self, x_columns, y_column, normalize=False):
		self._x_columns = x_columns
		self._y_column = y_column
		self._normalize = normalize
		self._untrained_model = None
		self._trained_model = None

	def fit(self, data):
		"""
		:type data: SparkDF or PandasDF
		:rtype: LogisticRegression
		"""






import numpy as np
import pandas as pd
from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame as SparkDF
from ...time.progress import ProgressBar
from ..wrangling._get_string_columns import get_string_columns


class OneHotEncoder:
	def __init__(
			self, top=10, rank_method='first', encode_na=False, replacement='other',
			include=None, exclude=None, lowercase=True
	):
		self._column_values = {}
		self._top = top
		self._method = rank_method
		self._encode_na = encode_na
		self._replacement = replacement
		self._include = include
		self._exclude = exclude
		self._lowercase = lowercase
		self._one_hot_columns = None
		self._encoded_columns = None

	@property
	def encoded_columns(self):
		"""
		:rtype: list[str]
		"""
		return self._encoded_columns

	@property
	def one_hot_columns(self):
		"""
		:rtype: list[str]
		"""
		return self._one_hot_columns

	def fit(self, data, echo=0):
		"""
		:type data: PandasDF or SparkDF
		:type echo: int or bool
		:rtype: OneHotEncoder
		"""
		echo = max(0, echo)
		if isinstance(data, PandasDF):
			self._pandas_fit(data=data, echo=echo)
		elif isinstance(data, SparkDF):
			self._spark_fit(data=data, echo=echo)
		else:
			raise TypeError(f'data of type {type(data)} is not supported!')
		return self

	def _get_encoding_columns(self, data):
		string_columns = get_string_columns(data=data)

		if self._include is not None:
			string_columns = [col for col in string_columns if col in self._include]
		if self._exclude is not None:
			string_columns = [col for col in string_columns if col not in self._exclude]

		return string_columns

	def _spark_fit(self, data, echo=0):
		"""
		:type data: SparkDF
		:type echo: bool or int
		:rtype: OneHotEncoder
		"""
		one_hot_columns = []
		columns_to_be_encoded = self._get_encoding_columns(data=data)

	def _pandas_fit(self, data, echo=0):
		"""
		:type data: PandasDF
		:type echo: bool or int
		:rtype: OneHotEncoder
		"""
		result = data.copy()
		one_hot_columns = []
		columns_to_be_encoded = self._get_encoding_columns(data=data)

		progress_bar = ProgressBar(total=len(columns_to_be_encoded))
		progress_amount = 0
		self._encoded_columns = []
		for col_name in columns_to_be_encoded:
			if echo:
				progress_bar.show(amount=progress_amount, text=f'DM training dummies for {col_name}')
			progress_amount += 1
			try:
				temp_data = data[[col_name]].copy()
				if self._lowercase:
					temp_data[col_name] = temp_data[col_name].str.lower()
				temp_data['count'] = 1
				counts = temp_data.groupby(col_name).sum().reset_index(drop=False)

				if self._top is not None:
					top_counts = counts.sort_values(by=col_name, ascending=False).head(self._top)
					only_include = set(top_counts[col_name])
					temp_data[col_name] = np.where(
						temp_data[col_name].isin(only_include), temp_data[col_name], self._replacement
					)

				else:
					only_include = set(counts[col_name])

				dummies = pd.get_dummies(
					data=temp_data[[col_name]],
					prefix=col_name, prefix_sep='_', dummy_na=self._encode_na, sparse=False
				)

				dummies = dummies[[col for col in dummies if not dummies[col].nunique() == 1]]

				result = pd.concat([result, dummies], axis=1)
				one_hot_columns += list(dummies.columns)

				self._column_values[col_name] = only_include
				self._encoded_columns.append(col_name)
			except Exception as e:
				print(f'exception being raised for column: {col_name}')
				raise e
		self._one_hot_columns = one_hot_columns
		if echo:
			progress_bar.show(amount=progress_amount, text=f'DM trained dummies for {self._encoded_columns}')
		return result

	def transform(self, data, echo=0):
		return self.encode(data=data, echo=echo)

	def fit_transform(self, data, echo=0):
		self.fit(data=data, echo=echo)
		return self.transform(data=data, echo=echo)

	def encode(self, data, drop_encoded=True, echo=0):
		echo = max(0, echo)
		result = data.copy()
		progress_bar = ProgressBar(total=len(self._column_values))
		progress_amount = 0
		for col_name, only_include in self._column_values.items():
			if echo:
				progress_bar.show(amount=progress_amount, text=f'DM creating dummies for {col_name}')
			progress_amount += 1
			temp_data = result[[col_name]].copy()
			if self._lowercase:
				temp_data[col_name] = temp_data[col_name].str.lower()
			temp_data[col_name] = np.where(
				temp_data[col_name].isin(only_include), temp_data[col_name], self._replacement
			)
			dummies = pd.get_dummies(
				data=temp_data[[col_name]],
				prefix=col_name, prefix_sep='_', dummy_na=self._encode_na, sparse=True
			)
			result = pd.concat([result, dummies], axis=1)
		for col_name in self._one_hot_columns:
			if col_name not in result.columns:
				result[col_name] = 0
		if echo:
			progress_bar.show(amount=progress_amount, text=f'DM created dummies for {self._encoded_columns}')

		if drop_encoded:
			result = result.drop(columns=self.encoded_columns)

		extra_columns = [
			x for x in result.columns
			if x not in self.encoded_columns + self.one_hot_columns and x not in data.columns
		]
		if len(extra_columns) > 0:
			result = result.drop(columns=extra_columns)

		return result

	def untransform(self, data, echo=0):
		echo = max(0, echo)
		result = data.copy()
		progress_bar = ProgressBar(total=len(self._column_values))
		progress_amount = 0
		# TODO

	@classmethod
	def get_encoder_and_encoded(cls, data, echo=0, **kwargs):
		"""
		:rtype: tuple[OneHotEncoder, PandasDF]
		"""
		one_hot_encoder = cls(**kwargs)
		data_encoded = one_hot_encoder.fit_transform(data=data, echo=echo)
		return one_hot_encoder, data_encoded


def get_one_hot_encoder_and_encoded(data, echo=0, **kwargs):
	return OneHotEncoder.get_encoder_and_encoded(data=data, echo=echo, **kwargs)



class TimeEstimate:
	def __init__(self, count=0, total=0):
		self._total = total
		self._count = count

	def append(self, elapsed):
		self._total += elapsed
		self._count += 1

	def get_mean(self):
		return self._total / self._count

	@property
	def mean(self):
		return self.get_mean()

	def __repr__(self):
		return str(self.mean)

	def __add__(self, other):
		"""
		:type other: TimeEstimate
		:rtype: TimeEstimate
		"""
		return TimeEstimate(count=self._count + other._count, total=self._total + other._total)


class MissingTimeEstimate:
	def __eq__(self, other):
		return isinstance(other, MissingTimeEstimate)

	def __ne__(self, other):
		return not isinstance(other, MissingTimeEstimate)



import multiprocess
import atexit
from time import sleep
from .BaseController import BaseController
from ._do_task import do_task
from ._DEFAULT_VALUES import *


class ProcessController(BaseController):
	def __init__(
			self,
			time_unit=TIME_UNIT,
			max_cpu_count=MAX_CPU_COUNT,
			sleep_time=SLEEP_TIME,
			empty_count_limit=EMPTY_COUNT_LIMIT,
			max_sleep_time=MAX_SLEEP_TIME
	):
		super().__init__(time_unit=time_unit)

		self._manager = multiprocess.Manager()

		# in parent class:
		self._to_do_queue = self._manager.Queue()
		self._done_queue = self._manager.Queue()
		self._tasks_status = self._manager.dict()
		self._incomplete_task_ids = self._manager.dict()

		# not in parent class:
		self._workers_doing = self._manager.dict()  # key: worker_id, value: task_id
		self._worker_reports = self._manager.dict()
		self._cpu_usage = self._manager.Value('i', 0)

		self._keep_workers_alive = self._manager.Value('i', 1)

		self._workers = dict()

		self._max_cpu_count = max_cpu_count or self.system_cpu_count
		self._sleep_time = sleep_time

		self._empty_count_limit = empty_count_limit
		self._max_sleep_time = max_sleep_time
		atexit.register(self.terminate)

	@property
	def workers_doing(self):
		"""
		:rtype: dict
		"""
		return self._workers_doing

	@property
	def to_do_count(self):
		return self.to_do_queue.qsize()

	@property
	def done_count(self):
		return self.done_queue.qsize()

	@property
	def being_done_count(self):
		return sum(1 for value in self._workers_doing.values() if value is not None)

	@property
	def worker_reports(self):
		"""
		:rtype: dict[str, WorkerReport]
		"""
		return self._worker_reports

	@property
	def tasks_status(self):
		"""
		:rtype: dict
		"""
		return dict(self._tasks_status)

	@property
	def cpu_usage(self):
		return self._cpu_usage.value

	def _add_task_to_to_do(self, task):
		self.to_do_queue.put(task)
		self._tasks_status[task.id] = 'added'
		self._task_counter += 1

	def keep_workers_alive(self):
		self._keep_workers_alive.value = 1

	def add_task(self, function, args=None, kwargs=None, task_id=None, cpu_count=1):
		self.keep_workers_alive()
		return super().add_task(function=function, args=args, kwargs=kwargs, task_id=task_id, cpu_count=1)

	def let_workers_die(self):
		self._keep_workers_alive.value = 0

	@property
	def workers(self):
		"""
		:rtype: dict[str, multiprocessing.Process]
		"""
		return self._workers

	@property
	def worker_count(self):
		return len(self._workers)

	def _generate_worker_id(self, prefix='worker'):
		return super()._generate_worker_id(prefix=prefix)

	def add_worker(self, echo=0):
		worker_id = self._generate_worker_id()
		if echo:
			print(f'adding worker {worker_id}')
		process = multiprocess.Process(
			target=do_task,
			kwargs={
				'to_do_queue': self._to_do_queue,
				'done_queue': self._done_queue,
				'workers_doing': self._workers_doing,
				'tasks_status': self._tasks_status,
				'worker_reports': self._worker_reports,
				'keep_workers_alive': self._keep_workers_alive,
				'worker_id': worker_id,
				'cpu_usage': self._cpu_usage,
				'max_cpu_count': self._max_cpu_count,
				'sleep_time': self._sleep_time,
				'empty_count_limit': self._empty_count_limit,
				'max_sleep_time': self._max_sleep_time,
				'echo': echo
			}
		)
		self._workers[worker_id] = process
		process.start()

	def _insufficient_workers(self):
		return self.cpu_usage < self._max_cpu_count and self.worker_count < self._max_cpu_count

	def add_workers_as_needed(self, echo=0):
		new_worker_count = 0
		while not self._to_do_queue.empty() and self._insufficient_workers():
			sleep(self._sleep_time)
			self.add_worker(echo=echo)
			new_worker_count += 1
		return new_worker_count

	def remove_dead_workers(self, echo=0):
		dead_count = 0
		live_workers = {}
		for worker_id, worker in self.workers.items():
			if worker.is_alive():
				live_workers[worker_id] = worker
			else:
				if echo:
					print(f'removing dead worker {worker_id}')
				if worker_id in self.workers_doing:
					incomplete_task_id = self.workers_doing[worker_id]
					if incomplete_task_id is not None:
						self.incomplete_task_ids[incomplete_task_id] = 1
				dead_count += 1
		self._workers = live_workers
		return dead_count

	def adjust_workers(self, echo=0):
		dead_count = self.remove_dead_workers()
		if self.to_do_count == 0:
			self.let_workers_die()
			new_worker_count = 0
		else:
			self.keep_workers_alive()
			new_worker_count = self.add_workers_as_needed(echo=echo)
		return {'dead': dead_count, 'new': new_worker_count}

	def terminate(self):
		terminated_count = 0
		for worker_id, worker in self.workers.items():
			worker.terminate()
			terminated_count += 1
		return terminated_count

	def _get_from_done_queue(self):
		return self.done_queue.get_nowait()

	def clean_up(self, echo=0):
		processed = self.process_done_queue(echo=echo)
		d = self.adjust_workers(echo=echo)
		dead_count = d['dead']
		new_count = d['new']
		return {'processed_tasks': len(processed), 'dead_workers': dead_count, 'new_workers': new_count}

	def get_status(self, return_text=True, return_values=False):
		todo = self.to_do_count
		being_done = self.being_done_count
		done = self.done_count
		processed = self.processed_count
		cpu_usage = self.cpu_usage
		worker_count = self.worker_count
		if return_values:
			result = {
				'to_do_count': todo,
				'being_done_count': being_done,
				'done_count': done,
				'processed_count': processed,
				'cpu_usage': cpu_usage,
				'worker_count': worker_count
			}
		else:
			result = None

		if return_text:
			text_list = [
				f'CPU: {cpu_usage}/{self._max_cpu_count}',
				f'Workers: {worker_count}',
				f'To-do: {todo}',
				f'Doing: {being_done}/{todo}',
				f'Done: {processed + done}',
				f'Processed: {processed}/{processed + done}        '
			]
			text = ' - '.join(text_list)
			if result is None:
				result = text
			else:
				result['text'] = text

		return result

	def show_status(self, clean_up=1, delay=1):
		try:
			while True:
				self.write(string=self.get_status(return_text=True, return_values=False))
				if clean_up:
					self.clean_up(echo=0)
				sleep(delay)
		except KeyboardInterrupt:
			self.write(string=self.get_status(return_text=True, return_values=False), flush=True)

	def wait_for_tasks(self, echo=1):
		try:
			while True:
				status = self.get_status(return_text=True, return_values=True)
				if echo:
					self.write(string=status['text'])
				d = self.clean_up(echo=0)
				sleep(self._sleep_time)
				if status['to_do_count'] + status['being_done_count'] == 0:
					if status['processed_count'] >= status['done_count']:
						break

		except KeyboardInterrupt:
			return 0

		self.clean_up(echo=0)
		status = self.get_status(return_text=True, return_values=True)
		if echo:
			self.write(string=status['text'] + '\n')

	def do(self, echo=1):
		self.wait_for_tasks(echo=echo)
		self.process_done_queue(echo=echo)



from ..wrangling import count_rows_without_missing
from ...collections import get_powerset


def find_columns_that_fully_cover_missing(data, column):
	"""
	if a column has missing values,
	find other columns such that these "helper columns" always have values
	for rows with missing values of the column in question
	"""
	rows_with_na = data[data[column].isna()]
	return [name for name, is_na in dict(rows_with_na.isna().any()).items() if not is_na]


def count_non_na_per_column(data):
	return dict(data.notna().sum())


def get_value_coverage(data, column):
	"""
	for a certain column with missing value,
	find how many values does every other column have for rows with missing values of the column
	"""
	non_na = data[data[column].notna()].drop(columns=column)
	return count_non_na_per_column(non_na)


def get_missing_coverage_per_column_subset(data, column):
	"""
	if a column has missing values, consider all subsets of other columns and see how much
	they cover
	"""
	viable_columns = find_columns_that_fully_cover_missing(data=data, column=column)
	non_empty_combinations = [l for l in get_powerset(iterable=viable_columns) if len(l) > 0]
	results = []
	for combination in non_empty_combinations:
		sub_data = data[combination]
		coverage = count_rows_without_missing(data=sub_data)
		column_ratio = len(combination)/(data.shape[1] - 1)
		row_ratio = coverage/data.shape[0]
		result = {
			'combination': combination,
			'row_coverage': coverage, 'row_ratio': row_ratio,
			'column_ratio': column_ratio,
			'row_squared_column_score': row_ratio ** 2 * column_ratio
		}
		results.append(result)
	return results


def find_best_coverage_combination(data, column, method='row_squared_column_score'):
	combinations = get_missing_coverage_per_column_subset(data=data, column=column)
	return sorted(combinations, key=lambda x: -x[method])[0]



from ._get_coefficients import get_coefficients
from ._get_feature_importances import get_feature_importances
from .InfluenceSimulator import InfluenceSimulator
from ._get_model_influence import get_model_influence


def get_coefficients(model, columns, raise_error=True, **kwargs):
	try:
		coef = model.coef_
		if isinstance(coef, dict):
			return coef
		try:
			coefficients = list(coef.flatten())
		except:
			coefficients = list(coef)
	except (AttributeError, KeyError):
		return None

	try:
		intercept = model.intercept_
	except AttributeError:
		intercept = None


	if len(columns) != len(coefficients):
		if raise_error:
			raise RuntimeError(f'number of columns: {len(columns)}, number of coefficients: {len(coefficients)}')
		else:
			return None
	else:
		coefficient_dictionary = {column: coefficient for coefficient, column in zip(coefficients, columns)}

		for key, value in kwargs.items():
			coefficient_dictionary[key] = value

		if intercept is not None and 'intercept' not in coefficient_dictionary:
			coefficient_dictionary['intercept'] = intercept

		return coefficient_dictionary



from pandas import DataFrame


class DataSlice:
	def __init__(self, data_id, indices, columns=None):
		self._data_id = data_id
		self._columns = columns
		self._indices = indices

	@property
	def data_id(self):
		return self._data_id

	def get_data(self, data):
		"""
		:type data: DataFrame
		:rtype: DataFrame
		"""
		if self._columns is not None:
			data = data[self._columns]
		if self._indices is not None:
			return data.iloc[self._indices]
		else:
			return data


class TrainingTestSlice:
	def __init__(self, data_id, training_indices, test_indices, columns=None):
		self._training_slice = DataSlice(data_id=data_id, indices=training_indices, columns=columns)
		self._test_slice = DataSlice(data_id=data_id, indices=test_indices, columns=columns)

	@property
	def data_id(self):
		return self._training_slice.data_id

	def get_training_data(self, data):
		return self._training_slice.get_data(data=data)

	def get_test_data(self, data):
		return self._test_slice.get_data(data=data)


from datetime import date


def convert_string_to_date(s):
	"""
	converts a date formated as YYYY-MM-DD to date
	:type s: str
	:rtype: date
	"""
	year = int(s[:4])
	month = int(s[5:7])
	day = int(s[8:10])
	return date(year, month, day)



from .date import get_date_part, yearmonth_to_date, get_monthname, get_weekday, get_weekdayname
from .date import get_quarter
from .date import get_month_num, get_month_name, get_months_list
from .date import add_days, add_weeks, add_months, add_quarters, add_years, add_time
from .date import get_today, get_today_str, get_date
from .date import YearMonth
from .date import get_number_of_days
from .date import convert_string_to_date
from .date import parse_date
from .date import get_week_number
from .date import get_date_mean

from .time import get_now, get_time
from .time import convert, get_elapsed, get_elapsed_months, get_elapsed_years
from .time import Timer
from .time import MeasurementSet
from .time import measure
from .estimate import Estimator



from .List import List
from ._flatten import flatten
from ._duplicates import get_duplicates, has_duplicates, remove_duplicates
from ._product import get_product
from .Counter import Counter
from ._make_hashable import make_hashable, is_hashable
from .OrderedSet import OrderedSet
from .cumsum import cumsum
from ._JupyterContainer import JupyterDictionary, JupyterList, JupyterSet, JupyterTuple, display_dataframe_container
from .Series import Series
from ._create_grid import create_grid
from ._cross_lists import cross_lists
from ._get_powerset import get_powerset, get_set_of_subsets
from .get_intersection import get_intersection



from ._sample_per_group import sample_per_group, shuffle
from .UpSampler import UpSampler
from ._sample_by_id import sample_by_id, split_by_id
from ._split_data import split_data
from ._balance_sample import balance_sample



from datetime import datetime
from .get_time import get_now
from .convert import convert


def _get_elapsed(start, end=None):
	"""
	:type start: datetime or date
	:type end: datetime or date or NoneType
	:rtype timedelta
	"""
	end = end or get_now()
	if not isinstance(end, datetime) and isinstance(start, datetime):
		start = start.date()
	elif isinstance(end, datetime) and not isinstance(start, datetime):
		end = end.date()
	return end - start


def get_elapsed_months(start, end=None):
	"""
	:type start: datetime
	:type end: datetime or NoneType
	:rtype: float
	"""
	end = end or get_now()
	return (end.year - start.year) * 12 + end.month - start.month + (end.day - start.day)/31


def get_elapsed_years(start, end=None):
	"""
	:type start: datetime
	:type end: datetime or NoneType
	:rtype: float
	"""
	return get_elapsed_months(start=start, end=end) / 12


def get_elapsed(start, end=None, unit='timedelta'):
	"""
	:param datetime or date start: start time
	:param datetime or or date or NoneType end: end time, the current time is used if not provided (None entered)
	:param str unit: 'timedelta', 'seconds', 'ms' (milliseconds), 'minutes', 'hours', 'days', 'months', or 'years'
	:rtype: float
	"""
	unit = unit.lower()
	u = unit[0]
	if u in ['t', 's', 'd', 'h'] or unit[:2] in ['ms', 'us'] or unit[:3] in ['min', 'mic', 'mil']:
		return convert(delta=_get_elapsed(start=start, end=end), to_unit=unit)
	elif unit[:2] == 'mo':  	# months
		return get_elapsed_months(start=start, end=end)
	elif u == 'y':
		return get_elapsed_years(start=start, end=end)
	else:
		raise ValueError(f'unit:{unit} is not recognizable.')


def find_best_unit(delta):
	seconds = convert(delta=delta, to_unit='s')
	if seconds < 1e-3:
		return 'microsecond'
	elif seconds < 1:
		return 'millisecond'
	elif seconds < 60:
		return 'second'
	elif seconds < 3600:
		return 'minute'
	elif seconds < 24 * 3600:
		return 'hour'
	else:
		return 'day'



from .SingleColumnImputer import SingleColumnImputer, ROW_NUM_COL
from pandas.api.types import is_numeric_dtype
from ...time.progress import ProgressBar
from pandas import DataFrame
from .exceptions import MissingModelError, IncludeExcludeClashError
from .exceptions import InputError
from copy import deepcopy


def get_columns_with_missing_values(data):
	"""
	:type data: DataFrame
	:rtype: list[str]
	"""
	return data.columns[data.isna().any()].tolist()


class Imputer:
	def __init__(
			self, regressor=None, classifier=None, impute_columns=None,
			exclude_columns=None, include_columns=None, echo=1
	):
		"""
		:param LinearRegression regressor: model to be used for imputation of numerical columns
		:param DecisionTreeClassifier or HyperModel classifier: model to be used for imputation of categorical columns
		:param NoneType or list[str] or str impute_columns: column(s) to be imputed, if None, all columns with missing values will be considered
		:param NoneType or list[str] or str exclude_columns: columns to be excluded from the model
		:param NoneType or list[str] or str include_columns: if not None, only these columns of the dataset will be used
		"""
		if regressor is None and classifier is None:
			raise MissingModelError('at least one of regressor or classifier should be provided!')

		self._regressor = regressor
		self._classifier = classifier

		if isinstance(impute_columns, str):
			impute_columns = [impute_columns]
		if isinstance(exclude_columns, str):
			exclude_columns = [exclude_columns]
		if isinstance(include_columns, str):
			include_columns = [include_columns]

		self._impute_columns = impute_columns
		self._exclude_columns = exclude_columns or []
		self._include_columns = include_columns

		if exclude_columns is not None and include_columns is not None:
			columns_in_both = [column for column in exclude_columns if column in include_columns]
			if len(columns_in_both) > 0:
				raise IncludeExcludeClashError(f'You cannot include and exclude a column! These are in both: {columns_in_both}')

		self._imputers = {}
		self._echo = echo
		self._logs = ''

	def fit(self, X):
		"""
		:param DataFrame X: the dataframe to be imputed
		:rtype: DataFrame
		"""
		# find columns that are only in X and are only numeric and are only in include
		ok_columns = [column for column in X.columns if is_numeric_dtype(X[column])]
		if self._include_columns is not None:
			ok_columns = [x for x in ok_columns if x in self._include_columns]
		if self._exclude_columns is not None:
			ok_columns = [x for x in ok_columns if x not in self._exclude_columns]

		impute_columns = self._impute_columns
		if impute_columns is None:
			impute_columns = get_columns_with_missing_values(data=X)
			if self._include_columns is not None:
				impute_columns = [col for col in impute_columns if col in self._include_columns]
			elif self._exclude_columns is not None:
				impute_columns = [col for col in impute_columns if col not in self._exclude_columns]

		if not isinstance(impute_columns, (list, str)) and impute_columns is not None:
			raise InputError(f'columns is of type {type(impute_columns)}')
		elif isinstance(impute_columns, str):
			impute_columns = [impute_columns]

		progress_bar = ProgressBar(total=len(impute_columns), echo=self._echo)
		progress = 0

		for column in impute_columns:
			progress_bar.show(amount=progress, text=f'fitting imputer for column: "{column}"')
			if is_numeric_dtype(X[column]):
				model = self._regressor
				column_type = 'numerical'
			else:
				model = self._classifier
				column_type = 'nonnumerical'

			if model is None:
				raise RuntimeError(f'no model for {column} which is {X[column].dtype}')

			model = deepcopy(model)
			single_column_imputer = SingleColumnImputer(
				estimator=model, column=column, column_type=column_type,
				use_columns=ok_columns
			)

			single_column_imputer.fit(X)
			self._imputers[column] = single_column_imputer

			model_name = model.__class__.__name__
			if model_name[0].lower() in ['a', 'e', 'i', 'o']:
				a = 'an'
			else:
				a = 'a'
			self._logs += f'fitted {a} "{model_name}" imputer for the {column_type} column "{column}".\n'
			progress += 1
		progress_bar.show(amount=progress, text=f'fitting imputers complete')
		self._logs += '\n'

	@property
	def single_column_imputers(self):
		"""
		:rtype: dict[str, SingleColumnImputer]
		"""
		return self._imputers

	def transform(self, X):
		"""
		:type X: DataFrame
		:rtype: DataFrame
		"""

		X = X.copy()
		X[ROW_NUM_COL] = range(X.shape[0])
		progress_bar = ProgressBar(total=len(self.single_column_imputers), echo=self._echo)
		progress = 0
		for column, imputer in self.single_column_imputers.items():
			progress_bar.show(amount=progress, text=f'imputating column "{column}"')
			X[column] = imputer.impute_column(data=X)
			self._logs += f'imputed column "{column}".\n'
			progress += 1
		progress_bar.show(amount=progress, text='imputatation complete')
		self._logs += '\n'
		return X.drop(columns=ROW_NUM_COL)

	def fit_transform(self, X):
		self.fit(X)
		return self.transform(X)



import re
from datetime import date as datetime_date
from calendar import day_name, day_abbr

from .add_time import add_months, add_days
from .YearMonth import YearMonth
from .YearQuarter import YearQuarter
from .get_quarter import get_quarter
from .remove_non_alphanumeric import remove_non_alphanumeric


def get_monthname(date, abr=False):
	try:

		result = date.strftime('%B')
	except:
		print('date:', date, type(date))
		raise ValueError('oops')
	if abr:
		return result[:3]
	else:
		return result


def get_weekday(date):
	return date.weekday()


def get_weekdayname(date, abr=False):
	weekday = get_weekday(date)
	if abr:
		return day_abbr[weekday]
	else:
		return day_name[weekday]


def get_date_part(date, date_part, sep='-', abr=False):
	date_part = date_part.lower()
	date_part = remove_non_alphanumeric(s=date_part, replace_with='', keep_underscore=False)
	if date_part == 'day':
		return date.day
	elif date_part == 'month':
		return date.month
	elif date_part == 'monthname':
		return get_monthname(date=date, abr=abr)
	elif date_part == 'quarter':
		return get_quarter(date=date)
	elif date_part == 'yearmonth':
		return YearMonth(date=date, sep=sep)
	elif date_part == 'yearmonthabr':
		return YearMonth(date=date, sep=sep, month_as='abr')
	elif date_part == 'yearmonthname':
		return YearMonth(date=date, sep=sep, month_as='name')
	elif date_part == 'yearquarter':
		return YearQuarter(date=date, sep=sep)
	elif date_part == 'monthname':
		return get_monthname(date=date, abr=abr)
	elif date_part == 'weekday':
		return get_weekday(date)
	elif date_part == 'weekdayname':
		return get_weekdayname(date=date, abr=abr)
	else:
		return date.year


def yearmonth_to_date(x, day=1):
	if type(x) is int:
		year = x // 100
		month = x % 100
	elif type(x) is str:
		x = str(x)
		year = int(re.findall(pattern=r'^\d+', string=x)[0][:4])
		month = int(re.findall(pattern=r'\d+$', string=x)[0][-2:])
	elif type(x) is YearMonth:
		year = x.year.value
		month = x.month.value
	else:
		return None

	try:
		result = datetime_date(year=year, month=month, day=day)
	except:
		# There is an exception if the day of the month we're in does not exist in the target month
		# Go to the FIRST of the month AFTER, then go back one day.
		first_day_of_the_month = datetime_date(year=year, month=month, day=1)
		first_day_of_next_month = add_months(date=first_day_of_the_month, months=1)
		last_day_of_the_month = add_days(date=first_day_of_next_month, days=-1)
		result = last_day_of_the_month

	return result



from collections import Counter
from ._make_hashable import make_hashable
from .Counter import Counter


def has_duplicates(l):
	"""
	returns True if list l has duplicates
	:type l: list
	:rtype: bool
	"""
	if len(l) == 0:
		return False
	else:
		return Counter(l).most_common()[0][1] > 1


def get_duplicates(l):
	"""
	returns duplicate values in list l
	:type l: list
	:rtype: list
	"""
	return l.__class__([k for k, v in Counter(l).items() if v > 1])


def remove_duplicates(l):
	"""
	removes the duplicates from a list l
	:type l: list
	:rtype: list
	"""
	used = set()
	unique_list = l.__class__([x for x in l if make_hashable(x) not in used and (used.add(make_hashable(x)) or True)])
	return unique_list
